classdef Conductance_CEM_electrode_parametrization < Functional_electrode_parametrization_interface
    properties
        Ne;
        electrode_conductances;
        conductance_bounds = [0.00001, 100000];
        electrode_domain; % Ne x 2. Start and end points on normalized electrode arc length
        mesh;
        precalc;
    end
    methods
        % obj = electrode object
        % gx = global x
        % gy = global y
        % electrode_ind = electrode index
        % electrode_l = electrode arc length parametrization (normalized)
        % bedge_l = single edge length parametrization = quadrature points = Nqp x 1
        % boundary_l = full boundary arc length parametrization (absolute)
        % RETURNS size(gx) array of conductances.
        function conds = conductances(obj, gx, gy, electrode_ind, electrode_l, bedge_l, boundary_l)
            conds = zeros(size(gx));
            elmask = logical(electrode_ind);
            conds(elmask) = obj.electrode_conductances(electrode_ind(elmask)) .* ...
                            (obj.electrode_domain(electrode_ind(elmask), 1) < electrode_l(elmask)) .* ...
                            (electrode_l(elmask) < obj.electrode_domain(electrode_ind(elmask), 2));
        end
        
        function obj = Conductance_CEM_electrode_parametrization(mesh)
            obj.Ne = max(mesh.bedgestream(:, 3));
            obj.electrode_domain = [zeros(obj.Ne, 1), ones(obj.Ne, 1)];
            obj.mesh = mesh;
            obj.precalc.Xb = [];
            obj.electrode_conductances = ones(obj.Ne, 1);
        end

        % Set parametrization so that each electrode has net contact
        % resistance of Rcs. Rcs may be a single value or a vector of N
        % elements where N is the number of electrodes
        function set_resistors(obj, Rcs)
            if length(Rcs) == 1
                Rcs = Rcs * ones(obj.Ne, 1);
            end

            [vf, idf] = unique(obj.mesh.bedgestream(:, 3), 'first');
            [vl, idl] = unique(obj.mesh.bedgestream(:, 3), 'last');
            [~, ordf] = sort(vf);
            [~, ordl] = sort(vl);
            idf = idf(ordf); idf = idf(2:end);
            idl = idl(ordl); idl = idl(2:end);
            ellens = (obj.electrode_domain(:, 2) - obj.electrode_domain(:, 1)) .* (obj.mesh.bedgestream(idl, 4) - obj.mesh.bedgestream(idf, 4));

            obj.electrode_conductances = [1./(Rcs.*ellens)];
        end

        % p = externally visible parameters. Must be consistent with
        % methods below.
        function p = get_params(obj)
            p = obj.electrode_conductances;
        end
        
        % Number of externally visible parameters. Must be consistent with
        % params, perturb(..) and derivative(..).
        function N = Nperturb(obj)
            N = obj.Ne;
        end

        % Perturb parametrization by parameter vector eta. Eta is in the
        % same units as derivatives generated by derivative(..) function.
        function perturb(obj, eta)
            obj.electrode_conductances = max(min(obj.electrode_conductances + eta, obj.conductance_bounds(2)), obj.conductance_bounds(1));
        end

        % Compute derivative for fixed pair (U-u)(V-v). I.e. compute the
        % integral \partial_t U(\sigma, \zeta(t), I; \eta) \cdot \tilde{I} = - \int_{\partial \Omega} \frac{\partial \zeta(t)}{\partial_t} \, (U - u) (\tilde{U} - \tilde{u}) \, {\rm d} S.
        % Xb = quadrature points (per mesh edge)
        % Wb = quadrature weights (per mesh edge)
        % vdrop2 = pointwise values of (U-u)(V-v) for the quadrature for fixed pairs (U,u) and (V,v). Nedges x length(Xb)
        % mesh = mesh
        function bp = boundary_params(obj, Xb, mesh)
            if (isequal(obj.precalc.Xb, Xb))
                bp = obj.precalc.bp;
                return
            end

            obj.precalc.bp = get_boundary_parametrization(mesh, Xb);
            bp = obj.precalc.bp;
        end

        function derivatives = derivative(obj, Xb, Wb, vdrop2, mesh)
            bp = obj.boundary_params(Xb, mesh);

            pointwise_d = -vdrop2 * Wb .* mesh.affine.detAb;
            derivatives = accumarray(bp.electrode_inds(logical(bp.electrode_inds(:, end)), end), pointwise_d(logical(bp.electrode_inds(:, end))));
        end
    end
end
