classdef Conductance_peak_electrode_parametrization < Functional_electrode_parametrization_interface
    properties
        Ne;
        params;
        limits;
        ellens;
    end
    properties (Hidden = true)
        precalc;
    end
    
    methods (Static = true)
        function val = beak(x, h, l, w)
            val = (x < l - w/2) .*              	0 + ...
                  (x >= (l - w/2)) .* (x < l) .* 	(2.*h.*(-2.*l + w + 2.*x))./(w.^2) + ...
                  (x >= l) .* (x < (l + w/2)) .* 	(2.*h.*(2.*l + w - 2.*x))./(w.^2) + ...
                  (x >= (l + w/2)) .*            	0;
        end

        function dh = beak_dh(x, h, l, w)
            dh = (x < l - w/2) .*               	0 + ...
                 (x >= (l - w/2)) .* (x < l) .* 	(2.*(-2.*l + w + 2.*x))./w.^2 + ...
                 (x >= l) .* (x < (l + w/2)) .* 	(2.*(2.*l + w - 2.*x))./w.^2 + ...
                 (x >= (l + w/2)) .*            	0;
        end

        function dl = beak_dl(x, h, l, w)
            dl = (x < l - w/2) .*               	0 + ...
                 (x >= (l - w/2)) .* (x < l) .* 	(-4.*h)./w.^2 + ...
                 (x >= l) .* (x < (l + w/2)) .* 	(4.*h)./w.^2 + ...
                 (x >= (l + w/2)) .*            	0;
        end

        function dw = beak_dw(x, h, l, w)
            dw = (x < l - w/2) .*               	0 + ...
                 (x >= (l - w/2)) .* (x < l) .*		(-2.*h.*(-4.*l + w + 4.*x))./w.^3 + ...
                 (x >= l) .* (x < (l + w/2)) .* 	(-2.*h.*(4.*l + w - 4.*x))./w.^3 + ...
                 (x >= (l + w/2)) .*           		0;
        end
    end
    
    methods
        function obj = Conductance_peak_electrode_parametrization(mesh)
            obj.Ne = max(mesh.bedgestream(:, 3));
            obj.params.h = ones(obj.Ne, 1);
            obj.params.l = 0.5 * ones(obj.Ne, 1);
            obj.params.w = 0.3 * ones(obj.Ne, 1);
            obj.limits.h = [10^-8, 10^8];
            obj.limits.l = [0.05, 0.95];
            obj.limits.w = [0.1, 1.0];

            [vf, idf] = unique(mesh.bedgestream(:, 3), 'first');
            [vl, idl] = unique(mesh.bedgestream(:, 3), 'last');
            [~, ordf] = sort(vf);
            [~, ordl] = sort(vl);
            idf = idf(ordf); idf = idf(2:end);
            idl = idl(ordl); idl = idl(2:end);
            obj.ellens = mesh.bedgestream(idl, 4) - mesh.bedgestream(idf, 4);
            obj.precalc.Xb = [];
            obj.precalc.derivatives = [];
            obj.precalc.params = [];
        end
        
        function set_resistors(obj, Rcs)
            if length(Rcs) == 1
                Rcs = Rcs * ones(obj.Ne, 1);
            end

            obj.params.h = (1./(Rcs(:).*obj.ellens)); % Conductance
%             obj.params.h = (Rcs(:).*obj.ellens); % Resistance
            obj.params.l = 0.5 * ones(obj.Ne, 1);
        end
        
        function conds = conductances(obj, gx, gy, electrode_ind, electrode_l, bedge_l, boundary_l)
            conds = zeros(size(gx));
            mask = logical(electrode_ind);
            conds(mask) = obj.beak(electrode_l(mask), obj.params.h(electrode_ind(mask)), obj.params.l(electrode_ind(mask)), obj.params.w(electrode_ind(mask)));
        end

        
        function p = get_params(obj)
            p = [obj.params.h; obj.params.l; obj.params.w];
        end
        
        function N = Nperturb(obj)
            N = obj.Ne * 3; % == [obj.params.h; obj.params.l; obj.params.w]
        end

        function perturb(obj, eta)
            obj.params.h = min(max(obj.params.h + eta(1:obj.Ne), ...
                                   obj.limits.h(1)), ...
                               obj.limits.h(2));
            obj.params.w = min(max(obj.params.w + eta(1+2*obj.Ne:end), obj.limits.w(1)), obj.limits.w(2));
            obj.params.l = min(max(obj.params.l + eta(1+1*obj.Ne:end-obj.Ne), obj.params.w/2), 1 - obj.params.w/2);
        end
        
        % "Flywheel" parameter derivative storage. Updates only when
        % changes have occured in obj.params or Xb.
        function param_derivatives = get_param_derivatives(obj, Xb, mesh)
            if (isequal(obj.precalc.Xb, Xb) && isequal(obj.precalc.params, obj.params))
                param_derivatives = obj.precalc.derivatives;
                return
            end
            obj.precalc.Xb = Xb;
            obj.precalc.params = obj.params;
            
            bp = get_boundary_parametrization(mesh, Xb);
            el_e_mask = logical(bp.electrode_inds);
            obj.precalc.derivatives.h = zeros(size(el_e_mask, 1), size(Xb, 2));
            obj.precalc.derivatives.l = zeros(size(el_e_mask, 1), size(Xb, 2));
            obj.precalc.derivatives.w = zeros(size(el_e_mask, 1), size(Xb, 2));
            obj.precalc.derivatives.bp = bp;

%             obj.precalc.derivatives.h(:) = obj.beak_dh(bp.electrode_l(el_e_mask), obj.params.h(bp.electrode_inds(el_e_mask)), obj.params.l(bp.electrode_inds(el_e_mask)), obj.params.w(bp.electrode_inds(el_e_mask))); % Per quadrature point
%             obj.precalc.derivatives.l(:) = obj.beak_dl(bp.electrode_l(el_e_mask), obj.params.h(bp.electrode_inds(el_e_mask)), obj.params.l(bp.electrode_inds(el_e_mask)), obj.params.w(bp.electrode_inds(el_e_mask))); % Per quadrature point
%             obj.precalc.derivatives.w(:) = obj.beak_dw(bp.electrode_l(el_e_mask), obj.params.h(bp.electrode_inds(el_e_mask)), obj.params.l(bp.electrode_inds(el_e_mask)), obj.params.w(bp.electrode_inds(el_e_mask))); % Per quadrature point

            temp = zeros(sum(el_e_mask(:, 1)), size(Xb, 2));
            temp(:) = obj.beak_dh(bp.electrode_l(el_e_mask), obj.params.h(bp.electrode_inds(el_e_mask)), obj.params.l(bp.electrode_inds(el_e_mask)), obj.params.w(bp.electrode_inds(el_e_mask))); % Per quadrature point
            obj.precalc.derivatives.h(el_e_mask) = temp(:);
            temp = obj.beak_dl(bp.electrode_l(el_e_mask), obj.params.h(bp.electrode_inds(el_e_mask)), obj.params.l(bp.electrode_inds(el_e_mask)), obj.params.w(bp.electrode_inds(el_e_mask))); % Per quadrature point
            obj.precalc.derivatives.l(el_e_mask) = temp(:);
            temp = obj.beak_dw(bp.electrode_l(el_e_mask), obj.params.h(bp.electrode_inds(el_e_mask)), obj.params.l(bp.electrode_inds(el_e_mask)), obj.params.w(bp.electrode_inds(el_e_mask))); % Per quadrature point
            obj.precalc.derivatives.w(el_e_mask) = temp(:);

            
            param_derivatives = obj.precalc.derivatives;
        end
        
        function derivatives = derivative(obj, Xb, Wb, vdrop2, mesh)
            param_derivatives = obj.get_param_derivatives(Xb, mesh);
            bp = param_derivatives.bp;

            el_e_mask = logical(bp.electrode_inds);
            
            vdrop2h = zeros(size(vdrop2));
            vdrop2l = zeros(size(vdrop2));
            vdrop2w = zeros(size(vdrop2));
            vdrop2h(:) = -vdrop2(:) .* param_derivatives.h(:) .* el_e_mask(:);
            vdrop2l(:) = -vdrop2(:) .* param_derivatives.l(:) .* el_e_mask(:);
            vdrop2w(:) = -vdrop2(:) .* param_derivatives.w(:) .* el_e_mask(:);
%             vdrop2h(el_e_mask) = -vdrop2(el_e_mask) .* param_derivatives.h(:);
%             vdrop2l(el_e_mask) = -vdrop2(el_e_mask) .* param_derivatives.l(:);
%             vdrop2w(el_e_mask) = -vdrop2(el_e_mask) .* param_derivatives.w(:);

            vh = vdrop2h * Wb .* mesh.affine.detAb;
            vl = vdrop2l * Wb .* mesh.affine.detAb;
            vw = vdrop2w * Wb .* mesh.affine.detAb;
            derivatives = [accumarray(bp.electrode_inds(logical(bp.electrode_inds(:, end)), end), vh(logical(bp.electrode_inds(:, end))));
                           accumarray(bp.electrode_inds(logical(bp.electrode_inds(:, end)), end), vl(logical(bp.electrode_inds(:, end))));
                           accumarray(bp.electrode_inds(logical(bp.electrode_inds(:, end)), end), vw(logical(bp.electrode_inds(:, end))))];
        end
    end
end