classdef Constant_log_domain_parametrization < Domain_parametrization_interface
    properties
        conductivity;
        mesh;
    end
    methods
        function obj = Constant_log_domain_parametrization(mesh)
            obj.conductivity = 1;
            obj.conductivity_bounds = [10^-6, 10^2];
            obj.mesh = mesh;
        end
        
        % obj = parametrization object
        % gx = global x
        % gy = global y
        % tx = reference triangle x
        % ty = reference triangle y
        % tid = point's triangle index
        % mesh = mesh structure. Should contained e.g. t2p matrix.
        % RETURNS size(gx) array of conductances.
        function conds = conductances(obj, gx, gy, tx, ty, tid, mesh)
            conds = obj.conductivity .* ones(size(gx));
        end

        % Set parametrization so that resulting conductance field is
        % constant.
        function set_constant(obj, val)
            obj.conductivity = val;
        end

        % p = externally visible parameters. Must be consistent with
        % methods below.
        function p = get_params(obj)
            p = log(obj.conductivity);
        end
        
        % p = externally visible parameters. Must be consistent with
        % methods below.
        function p = get_params_as_conductivity(obj)
            p = obj.conductivity;
        end

        % Number of externally visible parameters. Must be consistent with
        % params, perturb(..) and derivative(..).
        function N = Nperturb(obj)
            N = 1;
        end

        % Perturb parametrization by parameter vector eta. Eta is in the
        % same units as derivatives generated by derivative(..) function.
        function perturb(obj, eta)
            obj.conductivity = min(obj.conductivity_bounds(2), ...
                                   max(obj.conductivity_bounds(1), ...
                                       obj.conductivity .* exp(eta(:))));
        end


        % el_D = Elementwise derivative as N x N x Nt tensor
        function derivatives = derivative(obj, el_D, mesh)
            % Nodewise parametrization
            % Vval is a vector of 1/2 = value at sample points.
            % The "* (W/3)" is same as ".* (W')", as the sample
            % values are always the same. W/3 just provides only
            % one column which must be added twice separately for
            % each node in a triangle.
            % Could as well just take etaUfhatUghatEl ./ 6;
            %etaUfhatUghatV = (Vval.*(dLix.*dLjx + dLiy.*dLjy)) * (W/3).*abs(obj.mesh.affine.detA);
            %etaUfhatUghatV = reshape(etaUfhatUghatV, 1, 1, length(etaUfhatUghatV));
            el_D = obj.conductivity .* el_D;
            derivatives = accumndarray(el_D, (1:mesh.Nt), ones(1, mesh.Nt));
        end

        % Draw the parametrization in a way that makes or is correct sense.
        function P = draw_parametrization(obj, scaler)               
            X = obj.mesh.p(1,:);
            Y = obj.mesh.p(2,:);
            t = obj.mesh.t;

            P = patch(X(t), Y(t), ...
                      scaler(obj.conductivity) .* ones(size(X(t))), ...
                      scaler(obj.conductivity) .* ones(size(X(t))), ...
                      'FaceAlpha', 1.0, 'EdgeAlpha', 0.05);
            
            colorbar; colormap('jet');
            caxis(scaler([obj.conductivity_bounds(1) / 1.25, obj.conductivity_bounds(2) * 1.25]));
            axis square;
            if length(obj.mesh.t) > 5000
                set(P, 'edgecolor', 'none');
            end
        end
    end
end
