classdef Forward_model < matlab.mixin.Copyable
    properties
        mesh;
        domain_parametrization;
        electrode_model;
        system;
        loads;
        current_plan;
        solution; % canonical electrode basis
        domain_derivative_potentials; % canonical electrode basis
        gradient; % canonical electrode basis
        conductivity_system;
        solution_structure;
        perturbation_structure;
        conductivity_bounds = [0.01, 1];
        use_log = 0;
    end

    methods (Hidden = true)
        % Adds a dirichlet point to the system. Necessary to make the
        % system invertible.
        function postprocess_system_assembly(obj)
            obj.system = obj.conductivity_system;

            [~, min_index] = min(sum(obj.mesh.p.^2));
            obj.system(:, min_index) = 0;
            obj.system(min_index, :) = 0;
            obj.system(min_index, min_index) = 1;
        end

        % Shifts the solution to zero-mean (finite element!) space.
        function postprocess_solution(obj)
            obj.solution = [obj.solution(1:size(obj.mesh.p, 2), :) ...
                - sum(obj.solution(1:size(obj.mesh.p, 2), :) .* obj.mesh.affine.detp) ./ sum(obj.mesh.affine.detp);
                obj.solution(size(obj.mesh.p, 2)+1:end, :)];
            obj.domain_derivative_potentials = obj.solution;
        end

        function ss = system_size(obj)
            parts = fieldnames(obj.solution_structure);
            ss = [0 0];
            for partid=1:numel(parts)
                ss = ss + obj.solution_structure.(parts{partid});
            end
        end

        function el_conds = elementwise_conductivity(obj)
            X = [1/2 1/2 0 ; 0 1/2 1/2];
            W = [1/6 1/6 1/6]';
            gX{1} = bsxfun(@plus,obj.mesh.affine.Ax*X, obj.mesh.affine.bx);
            gX{2} = bsxfun(@plus,obj.mesh.affine.Ay*X, obj.mesh.affine.by);

            el_conds = obj.domain_parametrization.conductances(gX{1}, gX{2}, X(1, :), X(2, :), (1:obj.mesh.Nt)', obj.mesh);
            el_conds = mean(el_conds, 2); % Convert to size(Nt, 1)
        end

        function node_conds = nodewise_conductivity(obj)
            el_conds = obj.elementwise_conductivity();
            averager = repelem(1./sum(logical(obj.mesh.p2t'), 2), 1, size(obj.mesh.p2t', 2));
            node_conds = accumndarray(el_conds', obj.mesh.p2t', averager)';
        end
    end
    methods (Access = protected)
        function cp = copyElement(obj)
            % Make a shallow copy of all four properties
            cp = copyElement@matlab.mixin.Copyable(obj);
            % Make a deep copy of the DeepCp object
            cp.electrode_model = copy(obj.electrode_model);
            cp.domain_parametrization = copy(obj.domain_parametrization);
        end
    end
    methods
        function obj = Forward_model(mesh, domain_parametrization, electrodes, current_plan)
            obj.mesh = mesh;
            obj.current_plan = current_plan;
            obj.electrode_model = electrodes;
            obj.solution_structure.domain = [mesh.Np, mesh.Np];
            obj.domain_parametrization = domain_parametrization;
            obj.perturbation_structure.domain = [obj.domain_parametrization.Nperturb()];
        end

        function assemble_system(obj)
            X = [1/2 1/2 0 ; 0 1/2 1/2];
            W = [1/6 1/6 1/6]';
            
            % Define reference basisfunctions, their derivatives and
            % basiswise conductivity
            L{1} = 1-X(1,:)-X(2,:);
            L{2} = X(1,:);
            L{3} = X(2,:);

            Nip = size(X,2);

            dL{1} = [ -ones(1,Nip); -ones(1,Nip) ];
            dL{2} = [  ones(1,Nip); zeros(1,Nip) ];
            dL{3} = [ zeros(1,Nip);  ones(1,Nip) ];

%             A_weights = [0.5, 0.0, 0.5;
%                          0.5, 0.5, 0.0;
%                          0.0, 0.5, 0.5];
%             Vval = obj.elementwise_conductivity(obj.mesh.t)' * A_weights;

            Aval = obj.elementwise_conductivity();

            iind = [];
            jind = [];
            kk = zeros(9,length(obj.mesh.t));
            mind = 1;

            for i=1:3
                dLix = obj.mesh.affine.Px*dL{i};
                dLiy = obj.mesh.affine.Py*dL{i};

                for j=1:3
                    dLjx = obj.mesh.affine.Px*dL{j};
                    dLjy = obj.mesh.affine.Py*dL{j};

                    % Keep track of indices 
                    iind = [iind i]; jind = [jind j];
                    kk(mind, :) = (Aval.*(dLix.*dLjx + dLiy.*dLjy)) *W.*abs(obj.mesh.affine.detA);
                    mind = mind+1;
                end
            end

            system_size = obj.system_size();
            obj.conductivity_system = sparse(reshape(obj.mesh.t(iind,:), [], 1), ...
                                             reshape(obj.mesh.t(jind,:), [], 1), ...
                                             reshape(kk, [], 1), system_size(1), system_size(2));
                                         
            obj.postprocess_system_assembly();
        end

        function assemble_loads(obj)
            obj.loads = obj.electrode_model.discretized_load(obj.mesh);
        end

        function perturb(obj, eta)
            %obj.elementwise_conductivity(:) = min(obj.conductivity_bounds(2), max(obj.conductivity_bounds(1), obj.elementwise_conductivity + eta(1:obj.perturbation_structure.domain)));
            obj.domain_parametrization.perturb(eta(1:obj.domain_parametrization.Nperturb));
        end

        function solve_derivatives(obj)
            if (isempty(obj.domain_derivative_potentials))
                obj.solve_potentials();
            end

            unit_conductivity = Functional_conductivity(@(x, y) ones(size(x)));
            cond = unit_conductivity.sample_mesh(obj.mesh);

            % Changing these will break node parametrization!
            X = [1/2 1/2 0 ; 0 1/2 1/2];
            W = [1/6 1/6 1/6]';

            % Define reference basisfunctions, their derivatives and
            % basiswise conductivity
            Nt = obj.mesh.Nt;
            Np = obj.mesh.Np;
            Nip = size(X,2);
            dL{1} = [ -ones(1,Nip); -ones(1,Nip) ];
            dL{2} = [  ones(1,Nip); zeros(1,Nip) ];
            dL{3} = [ zeros(1,Nip);  ones(1,Nip) ];

            Aval = cond;
            Vval = 1/2 * cond;

            Ncurrents = size(obj.loads, 2);
            metaUfUgEl = zeros(Ncurrents, Ncurrents, Nt);

            for i=1:3
                dLix = obj.mesh.affine.Px*dL{i};
                dLiy = obj.mesh.affine.Py*dL{i};

                fpotentials = obj.domain_derivative_potentials(obj.mesh.t(i, :), :)';
                fpotentials = reshape(fpotentials, size(fpotentials, 1), 1, size(fpotentials, 2));

                for j=1:3
                    gpotentials = obj.domain_derivative_potentials(obj.mesh.t(j, :), :)';
                    gpotentials = reshape(gpotentials, 1, size(gpotentials, 1), size(gpotentials, 2));

                    dLjx = obj.mesh.affine.Px*dL{j};
                    dLjy = obj.mesh.affine.Py*dL{j};

                    % rows vary input currents, columns vary voltage
                    % measurements, pages vary conductivity

                    % Elementwise parametrization
                    etaUfhatUghatEl = (Aval.*(dLix.*dLjx + dLiy.*dLjy)) * W.*abs(obj.mesh.affine.detA);
                    etaUfhatUghatEl = reshape(etaUfhatUghatEl, 1, 1, length(etaUfhatUghatEl));
                    metaUfUgEl(:, :, :) = metaUfUgEl - fpotentials .* gpotentials .* etaUfhatUghatEl;
                end
            end
            % Nodewise parametrization
            % Vval is a vector of 1/2 = value at sample points.
            % The "* (W/3)" is same as ".* (W')", as the sample
            % values are always the same. W/3 just provides only
            % one column which must be added twice separately for
            % each node in a triangle.
            % Could as well just take etaUfhatUghatEl ./ 6;
            %etaUfhatUghatV = (Vval.*(dLix.*dLjx + dLiy.*dLjy)) * (W/3).*abs(obj.mesh.affine.detA);
            %etaUfhatUghatV = reshape(etaUfhatUghatV, 1, 1, length(etaUfhatUghatV));

            gradients = obj.domain_parametrization.derivative(metaUfUgEl, obj.mesh);

            obj.gradient = gradients;
        end

        function solve_potentials(obj)
            if (isempty(obj.system))
                obj.assemble_system();
            end
            if (isempty(obj.electrode_model))
                error('No current patterns defined for forward model.')
            end
            if (isempty(obj.loads))
                obj.assemble_loads();
            end
            
            obj.solution = obj.system \ obj.loads;
            obj.postprocess_solution;
        end

        function P = draw_potential_field(obj, solution_id)
            if (nargin == 1)
                solution_id = 1;
            end
            X = obj.mesh.p(1,:);
            Y = obj.mesh.p(2,:);
            t = obj.mesh.t;

            hold on;
            sol_in_el_basis = obj.electrode_model.basis_matrix() \ obj.current_plan(:, solution_id);
            P = patch(X(t), Y(t), ...
                      reshape(obj.solution(t, :) * sol_in_el_basis, size(X(t))), ...
                      reshape(obj.solution(t, :) * sol_in_el_basis, size(X(t))));
            %axis square;
            if length(obj.mesh.t) > 5000
                set(P, 'edgecolor', 'none');
            end

            colorbar; grid on; view(-20, 5);
        end

        function scaler = get_scaler(obj)
            if obj.use_log == 0
                scaler = @(x) x;
            elseif obj.use_log == 1
                scaler = @(x) log10(x);
            elseif obj.use_log == 2
                scaler = @(x) log10(log10(x));
            end
        end

        function P = draw_conductivity_field(obj, varargin)
            scaler = obj.get_scaler();
            hold on;
            if (ismethod(obj.domain_parametrization, 'draw_parametrization'))
                obj.domain_parametrization.draw_parametrization(scaler);
            else
                X = obj.mesh.p(1,:);
                Y = obj.mesh.p(2,:);
                t = obj.mesh.t;

                el_conds = obj.elementwise_conductivity();

                P = patch(X(t), Y(t), ...
                          scaler(repelem(el_conds', 3, 1)), ...
                          scaler(repelem(el_conds', 3, 1)), ...
                          'FaceAlpha', 1.0, 'EdgeAlpha', 0.05, varargin{:});
                          %'FaceColor', 'flat', 'FaceVertexCData', ...
                          %repmat(obj.elementwise_conductivity(t)', 1, 3));
                %salpha(1);
                colorbar; colormap('jet');
                caxis(scaler([obj.domain_parametrization.conductivity_bounds(1) / 1.25, obj.domain_parametrization.conductivity_bounds(2) * 1.25]));
                axis square;
                if length(obj.mesh.t) > 5000
                    set(P, 'edgecolor', 'none');
                end
            end
        end

        function P = draw_mesh(obj)
            X = obj.mesh.p(1,:);
            Y = obj.mesh.p(2,:);
            t = obj.mesh.t;

            hold on;
            P = patch(X(t), Y(t), ...
                      ones(size(X(t))));
            colorbar; colormap('gray');
            axis square;
            if length(obj.mesh.t) > 5000
                set(P, 'edgecolor', 'none');
            end
        end

        % Electrode voltages in Euclidean basis for current_plan currents
        function volts = get_NtoD(obj)
            if isempty(obj.solution)
                obj.solve_potentials();
            end
            volts = obj.electrode_model.basis_matrix * obj.electrode_model.electrode_potentials(obj.solution) * (obj.electrode_model.basis_matrix \ obj.current_plan);
        end

        % Gradient in Euclidean basis
        function grad = get_gradient(obj)
            if isempty(obj.gradient)
                obj.solve_derivatives();
            end
            
            left = obj.electrode_model.basis_matrix();
            right = obj.electrode_model.basis_matrix() \ obj.current_plan;
            grad = mtimesx(obj.gradient, right);
            grad = mtimesx(left, grad);
        end

        % Domain potentials with current_plan currents
        function volts = get_domain_potentials(obj)
            if isempty(obj.solution)
                obj.solve_potentials();
            end
            volts = obj.solution * (obj.electrode_model.basis_matrix() \ obj.current_plan);
            volts = volts(1:obj.solution_structure.domain, :);
        end
    end
end
