classdef Functional_electrodes < matlab.mixin.Copyable
    properties
        parametrization;
        mesh;
        current_basis;
        Ne;
        Nperturb;
    end

    methods (Static = true)
        function zm = zeromeaner(Ne)
            zm = zeromeangramschmidt(zeromeanpattern(Ne - 1));
            zm(abs(zm) < 10^-10) = 0;
        end
    end
    
    methods (Access = protected)
        function cp = copyElement(obj)
            parcopy = copy(obj.parametrization);
            cp = Functional_electrodes(obj.mesh, parcopy);
        end
    end

    methods
        function obj = Functional_electrodes(mesh, parametrization)
            obj.mesh = mesh;
            obj.parametrization = parametrization;
            obj.current_basis = Functional_electrodes.zeromeaner(max(mesh.bedgestream(:, 3)));
            obj.Nperturb = obj.parametrization.Nperturb();
            obj.Ne = max(mesh.bedgestream(:, 3));
        end
        
        function set_resistors(obj, Rcs)
            obj.parametrization.set_resistors(Rcs);
        end
        
        function system_appendix_triplets = system_appendix(obj, mesh)
            Xb = [(1-1/sqrt(3))/2  (1+1/sqrt(3))/2];
            Wb = [0.5 0.5]';
            % High order Simpson, because nothing else seems to get
            % sufficient precission for e.g. CEM. The jump in conductivity
            % quickly leads to significant (>> 1%) errors in solutions. If
            % you cheat with precisely known electrode locations and fit
            % your mesh to them, you can get away with 2nd order quadrature
            % (above).
            Xb = (0:1/128:1);
            Wb = [1, 4, repmat([2 4], 1, (length(Xb)-3)/2), 1]' * 1/(length(Xb)-1) ./ 3;

            Nibp = size(Xb, 2);

            Lb{1} = fliplr(Xb);
            Lb{2} = Xb;
            dLb{1} = [-ones(1, Nibp)];
            dLb{2} = [ones(1, Nibp)];            

            bp = get_boundary_parametrization(obj.mesh, Xb);
            conductance = obj.parametrization.conductances(bp.gXb{1}, bp.gXb{2}, bp.electrode_inds, bp.electrode_l, bp.Xb, bp.boundary_l); % Per quadrature point

            triplets.rows = [];
            triplets.cols = [];
            triplets.vals = [];

            %  _____________
            % |        |    |
            % |        |    |
            % |   uv   | uV |
            % |        |    |
            % |________|____|
            % |        |    |
            % |   Uv   | UV |
            % |________|____|
            %

            % \int_{e_l} z(x) (u-U)(v-V) dS
            % = \int_{e_l} z(x) uv - uV - Uv + UV dS

            % Indexing:
            % mesh.bedges <=> u, v
            % mesh.Np + K <=> U, V
            % Note: mesh.affine.(*)b corresponds to mesh.bedges
            
            % uv
            uvval = [];
            uviind = [];
            uvjind = [];
            for i = 1:length(Lb)
                Lbi = Lb{i};
                for j = 1:length(Lb)
                    Lbj = Lb{j};
                    uvval = [uvval; [(conductance .* Lbi .* Lbj .* mesh.affine.detAb) * Wb]'];
                    uviind = [uviind; i];
                    uvjind = [uvjind; j];
                end
            end

            triplets.rows = coly(mesh.bedges(uviind, :));
            triplets.cols = coly(mesh.bedges(uvjind, :));
            triplets.vals = coly(uvval);

            
            % uV / Uv
            Uvval = [];
            uviind = [];
            uvjind = [];
            for i = 1:length(Lb)
                Lbi = Lb{i};
                Uvval = [Uvval; [conductance .* Lbi .* 1.0 .* mesh.affine.detAb * Wb]'];
            end

            Uvblock = -accumarray([coly(mesh.bedges(:, bp.elmask')'), coly(repmat(bp.electrode_inds(bp.elmask, 1), 1, 2))], coly(Uvval(:, bp.elmask')'), [], [], [], true);
            [Uv_rows, Uv_cols, Uv_vals] = find(Uvblock * obj.current_basis);
            
            triplets.rows = [triplets.rows; Uv_rows];
            triplets.cols = [triplets.cols; Uv_cols + mesh.Np];
            triplets.vals = [triplets.vals; Uv_vals];
            
            triplets.rows = [triplets.rows; Uv_cols + mesh.Np];
            triplets.cols = [triplets.cols; Uv_rows];
            triplets.vals = [triplets.vals; Uv_vals];
            
            
            % UV
            UVval = [(conductance .* 1.0 .* 1.0 .* mesh.affine.detAb) * Wb];
            UVval = accumarray(bp.electrode_inds(bp.elmask, 1), UVval(bp.elmask));
            UVval = (obj.current_basis .* UVval)' * obj.current_basis;
            [UV_rows, UV_cols, UV_vals] = find(UVval);
            triplets.rows = [triplets.rows; UV_rows + mesh.Np];
            triplets.cols = [triplets.cols; UV_cols + mesh.Np];
            triplets.vals = [triplets.vals; UV_vals];

            system_appendix_triplets = triplets;
        end
        
        function derivatives = electrode_derivatives(obj, solution, mesh)
            derivatives = zeros(obj.Ne-1, obj.Ne-1, obj.parametrization.Nperturb());
%             Xb = (0:1/128:1);
            Xb = (0:1/16:1);
            Wb = [1, 4, repmat([2 4], 1, (length(Xb)-3)/2), 1]' * 1/(length(Xb)-1) ./ 3;

            Lb{1} = fliplr(Xb);
            Lb{2} = Xb;

            el_p_mask = logical(mesh.bedgestream(:, 3));

            for alpha = 1:size(obj.basis_matrix, 2)
                for beta = 1:size(obj.basis_matrix, 2)
                    ualpha = solution(1:end-obj.Ne+1, alpha);
                    ubeta = solution(1:end-obj.Ne+1, beta);
                    Ualpha = obj.basis_matrix() * solution(end-obj.Ne+2:end, alpha);
                    Ubeta = obj.basis_matrix() * solution(end-obj.Ne+2:end, beta);

                    % Numerical boundary integral
                    uas = ualpha(mesh.bedgestream(:, 1)) .* Lb{1} + ualpha(circshift(mesh.bedgestream(:, 1), [-1, 0])) .* Lb{2};
                    ubs = ubeta(mesh.bedgestream(:, 1)) .* Lb{1} + ubeta(circshift(mesh.bedgestream(:, 1), [-1, 0])) .* Lb{2};
                    Uas = zeros(size(mesh.bedgestream(:, 1)));
                    Ubs = zeros(size(mesh.bedgestream(:, 1)));
                    Uas(el_p_mask) = Ualpha(mesh.bedgestream(el_p_mask, 3));
                    Ubs(el_p_mask) = Ubeta(mesh.bedgestream(el_p_mask, 3));

                    vdrop2 = (Uas - uas).*(Ubs - ubs);

                    derivatives(alpha, beta, :) = obj.parametrization.derivative(Xb, Wb, vdrop2, mesh);
                end
            end

        end
        
        function perturb(obj, eta)
            obj.parametrization.perturb(eta);
        end
        
        function loads = discretized_load(obj, mesh)
            currents = eye(size(obj.current_basis, 2));

            loads = [zeros(mesh.Np, size(currents, 2));
                     currents];
        end
        
        function M = basis_matrix(obj)
            M = obj.current_basis;
        end
        
        function M = injection_basis_matrix(obj)
            M = obj.current_basis;
        end

        function potentials = electrode_potentials(obj, solution)
            potentials = solution(end-obj.Ne+2:end, :);
        end
        
        function label_electrodes(obj, solution_id, mesh, solution, scaler)
            if nargin == 1
                solution_id = 1;
            end
            if nargin < 4
                mesh = evalin('base', 'mesh');
                solution = evalin('base', 'fm.get_domain_potentials()');
            end
            if nargin < 5
                scaler = @(x) x;
            end
            hold on;

            [electrodes, IAbegin] = unique(mesh.bedgestream(:, 3), 'first');
            [~, IAend] = unique(mesh.bedgestream(:, 3), 'last');
            IAbegin = IAbegin(electrodes ~= 0);
            IAend = IAend(electrodes ~= 0);
            electrodes = electrodes(electrodes ~= 0);
            midpoints = (mesh.bedgestream(IAend, 4) - mesh.bedgestream(IAbegin, 4))./2 + mesh.bedgestream(IAbegin, 4);
            vq = interp1(mesh.bedgestream(:, 4), [mesh.p(:, mesh.bedgestream(:, 1))', scaler(solution(mesh.bedgestream(:, 1), solution_id))], midpoints);

            for el_id_id = 1:obj.Ne
                el_id = electrodes(el_id_id);
                plot3(vq(el_id, 1), vq(el_id, 2), vq(el_id, 3), '*', 'color', [0.5, 0.5, 0.5]);
                text(vq(el_id, 1), vq(el_id, 2), vq(el_id, 3), sprintf("%d", el_id), 'color', [0.5, 0.5, 0.5], 'FontSize', 15);
            end
        end
        
        function h = plot_electrodes_conductances(obj, electrode_ids, scaler, varargin)
            Xb = (0:1/32:1);

            bp = get_boundary_parametrization(obj.mesh, Xb);
            conductance = obj.parametrization.conductances(bp.gXb{1}, bp.gXb{2}, bp.electrode_inds, bp.electrode_l, bp.Xb, bp.boundary_l); % Per quadrature point

            if (nargin == 1)
                electrode_ids = [1];
            end
            
            electrode_l = bp.electrode_l';
            electrode_inds = bp.electrode_inds';
            conductance = conductance';
            for electrode_id = electrode_ids
                l = electrode_l(electrode_inds == electrode_id);
                val = conductance(electrode_inds == electrode_id);
                max_val = max(max(val));
                if nargout == 1
                    h = myplot(l(:), max(val(:)*scaler, max_val/10^5*scaler), varargin{:}); hold on;
                else
                    myplot(l(:), max(val(:)*scaler, max_val/10^5*scaler), varargin{:}); hold on;
                end
            end
        end
        
        function [integrals, midpoints] = conductance_integrals(obj)
            Xb = (0:1/128:1);
            Wb = [1, 4, repmat([2 4], 1, (length(Xb)-3)/2), 1]' * 1/(length(Xb)-1) ./ 3;

            Lb{1} = fliplr(Xb);
            Lb{2} = Xb;

            bp = get_boundary_parametrization(obj.mesh, Xb);
            conductance = obj.parametrization.conductances(bp.gXb{1}, bp.gXb{2}, bp.electrode_inds, bp.electrode_l, bp.Xb, bp.boundary_l); % Per quadrature point

            elemval = [];
            iind = [];
            for i = 1:length(Lb)
                Lbi = Lb{i};
                elemval = [elemval; [(conductance .* Lbi .* obj.mesh.affine.detAb) * Wb]'];
                iind = [iind; i];
            end
            
            ps = coly(obj.mesh.bedges(iind, :));
            vals = coly(elemval);
            
            p_to_elid = zeros(obj.mesh.Np, 1);
            p_to_elid(obj.mesh.bedgestream(:, 1)) = obj.mesh.bedgestream(:, 3);
            p_to_elid(p_to_elid == 0) = obj.Ne + 1;
            
            integrals = accumarray(p_to_elid(ps), vals);
            integrals = integrals(1:obj.Ne);



            midpoints = [];
            wconductance = conductance .* obj.mesh.affine.detAb .* Wb';
            for eli = 1:obj.Ne
                el_wconductance = coly(wconductance(obj.mesh.bedgestream(:, 3) == eli, :)');
                el_partial_conductance = cumsum(el_wconductance);
                midpoint_id = find(el_partial_conductance > el_partial_conductance(end)/2, 1, 'first');
                el_l = bp.electrode_l(obj.mesh.bedgestream(:, 3) == eli, :);
                [col, row] = ind2sub(size(el_l'), midpoint_id);
                %fprintf("%d - integral = %f, half = %f, l = %f and cond_l = %f\n", eli, integrals(eli), integrals(eli) / 2, el_l(row, col), el_partial_conductance(midpoint_id));
                midpoints(end+1, 1) = el_l(row, col);
            end
        end
    end
end