classdef Nodewise_conductivity_domain_parametrization < Domain_parametrization_interface
    properties
        nodewise_conductivity;
        mesh;
    end
    methods
        function obj = Nodewise_conductivity_domain_parametrization(mesh)
            obj.nodewise_conductivity = 0.1 * ones(mesh.Np, 1);
            obj.conductivity_bounds = [10^-2, 10^2];
            obj.mesh = mesh;
        end
        
        % obj = parametrization object
        % gx = global x
        % gy = global y
        % tx = reference triangle x
        % ty = reference triangle y
        % tid = point's triangle index
        % mesh = mesh structure. Should contained e.g. t2p matrix.
        % RETURNS size(gx) array of conductances.
        function conds = conductances(obj, gx, gy, tx, ty, tid, mesh)
            if not(isequal([1/2 1/2 0], tx) && isequal([0 1/2 1/2], ty))
                error("Nodewise domain parametrization expects quadrature with X = [1/2 1/2 0 ; 0 1/2 1/2] and W = [1/6 1/6 1/6]'.");
            end

            A_weights = [0.5, 0.0, 0.5;
                         0.5, 0.5, 0.0;
                         0.0, 0.5, 0.5];
            conds = mean(obj.nodewise_conductivity(mesh.t)', 2);
        end

        % Set parametrization so that resulting conductance field is
        % constant.
        function set_constant(obj, val)
            obj.nodewise_conductivity(:) = val;
        end

        % p = externally visible parameters. Must be consistent with
        % methods below.
        function p = get_params(obj)
            p = obj.nodewise_conductivity;
        end

        % Number of externally visible parameters. Must be consistent with
        % params, perturb(..) and derivative(..).
        function N = Nperturb(obj)
            N = size(obj.nodewise_conductivity, 1);
        end

        % Perturb parametrization by parameter vector eta. Eta is in the
        % same units as derivatives generated by derivative(..) function.
        function perturb(obj, eta)
            obj.nodewise_conductivity = min(obj.conductivity_bounds(2), ...
                                            max(obj.conductivity_bounds(1), ...
                                                obj.nodewise_conductivity + eta));
        end

        
        % el_D = Elementwise derivative as N x N x Nt tensor
        function derivatives = derivative(obj, el_D, mesh)
            % Nodewise parametrization
            % Vval is a vector of 1/2 = value at sample points.
            % The "* (W/3)" is same as ".* (W')", as the sample
            % values are always the same. W/3 just provides only
            % one column which must be added twice separately for
            % each node in a triangle.
            % Could as well just take etaUfhatUghatEl ./ 6;
            %etaUfhatUghatV = (Vval.*(dLix.*dLjx + dLiy.*dLjy)) * (W/3).*abs(obj.mesh.affine.detA);
            %etaUfhatUghatV = reshape(etaUfhatUghatV, 1, 1, length(etaUfhatUghatV));
            derivatives = accumndarray(el_D, mesh.p2t', 1./3 * ones(size(mesh.p2t')));
        end
        
        % Draw the parametrization in a way that makes or is correct sense.
        function P = draw_parametrization(obj, scaler)               
            X = obj.mesh.p(1,:);
            Y = obj.mesh.p(2,:);
            t = obj.mesh.t;

            P = patch(X(t), Y(t), ...
                      scaler(obj.nodewise_conductivity(t)), ...
                      scaler(obj.nodewise_conductivity(t)), ...
                      'FaceAlpha', 1.0, 'EdgeAlpha', 0.05);
            
            colorbar; colormap('jet');
            caxis(scaler([obj.conductivity_bounds(1) / 1.25, obj.conductivity_bounds(2) * 1.25]));
            caxis(scaler([min(obj.nodewise_conductivity) / 1.25, max(obj.nodewise_conductivity) * 1.25]));
            axis square;
            if length(obj.mesh.t) > 5000
                set(P, 'edgecolor', 'none');
            end
        end
    end
end
