#include "mex.hpp"
#include "mexAdapter.hpp"
#include "MatlabDataArray/ArrayFactory.hpp"

#include <algorithm>
#include <numeric>
#include <iostream>

using namespace matlab::data;
using namespace matlab::mex;

class MexFunction : public matlab::mex::Function {
private:
	void errorMsg(std::string msg) {
		std::shared_ptr<matlab::engine::MATLABEngine> matlabPtr = getEngine();
		matlab::data::ArrayFactory factory;

		matlabPtr->feval(u"error", 0, std::vector<matlab::data::Array>({ factory.createScalar(msg)}));
	}

	template <typename T>
	size_t arrayMaxAsInd(const TypedArray<T>& arr)
	{
		return static_cast<size_t>(*std::max_element(arr.cbegin(), arr.cend()));
	}

	void checkArgs(ArgumentList outputs, ArgumentList inputs) {
		if (inputs.size() < 2) {
			errorMsg("Accumndarray(vals, inds, weights) takes 2 or 3 arguments.");
		}
		if (outputs.size() != 1) {
			errorMsg("Accumndarray(vals, inds, weights) returns exactly 1 output.");
		}

		const Array& vals_typetest = inputs[0];
		const Array& inds_typetest = inputs[1];


		if (vals_typetest.getType() != ArrayType::DOUBLE) {
			errorMsg("Expected double array for vals.");
		}
		if (inds_typetest.getType() != ArrayType::DOUBLE) {
			errorMsg("Expected double array for inds.");
		}

		const TypedArray<double> vals = inputs[0];
		const TypedArray<double> inds = inputs[1];
		ArrayDimensions val_dims = vals.getDimensions();

		size_t max_ind = arrayMaxAsInd(inds);

		if (*val_dims.crbegin() < max_ind) {
			errorMsg("max(inds) must be less or equal to last dimension of vals.");
		}

		if (inputs.size() == 3) {
			const Array& weights = inputs[2];
			if (inds.getDimensions() != weights.getDimensions()) {
				errorMsg("inds and weights must have the same dimensions in Accumndarray.");
			}
			if (weights.getType() != ArrayType::DOUBLE) {
				errorMsg("Expected double array for weights.");
			}
		}

		if (outputs.size() > 1) {
			errorMsg("Accumndarray returns only one output.");
		}
	}

	TypedArray<uint32_t> doubleArrayToUints(const TypedArray<double>& arr) {
		ArrayFactory factory;

		TypedArray<uint32_t> out = factory.createArray<uint32_t>(arr.getDimensions());
		auto out_iter = out.begin();
		auto arr_iter = arr.cbegin();
		for (size_t i = 0; i < out.getNumberOfElements(); i++) {
			*out_iter = static_cast<uint32_t>(*arr_iter);
			if (static_cast<double>(*out_iter) != *arr_iter) {
				errorMsg("Casting error. Expected double array of integers.");
			}
			out_iter++;
			arr_iter++;
		}
		return out;
	}

	void accumulateBlock(double* target, const double* source,
	                     const double weight, const size_t block_size)
	{
		for (size_t i = 0; i < block_size; i++) {
			target[i] = target[i] + weight * source[i];
		}
	}

	TypedArray<double> indexedAccumulate(const TypedArray<double>& values,
	                                     const TypedArray<uint32_t>& indices,
	                                     const TypedArray<double>& weights)
	{
		ArrayFactory factory;

		const ArrayDimensions val_dims = values.getDimensions();
		const ArrayDimensions ind_dims = indices.getDimensions();
		ArrayDimensions out_dims(val_dims.cbegin(), std::prev(val_dims.cend()));
		out_dims.push_back(*ind_dims.cbegin());

		auto prod_op = [](const uint32_t& a, const uint32_t& b) {return a * b;};
		const uint32_t block_size = std::accumulate(val_dims.cbegin(),
		                                            std::prev(val_dims.cend()),
		                                            1, prod_op);


		size_t out_size = block_size * (*ind_dims.cbegin());

		buffer_ptr_t<double> out_buf = factory.createBuffer<double>(out_size);
		for (size_t i = 0; i < out_size; i++) {
			out_buf[i] = 0.0;
		}

		auto val_buf = values.begin();

		for (size_t ind_row = 0; ind_row < ind_dims[0]; ind_row++) {
			for (size_t ind_col = 0; ind_col < ind_dims[1]; ind_col++) {
				if (indices[ind_row][ind_col] != 0) {
					accumulateBlock(&out_buf[ind_row * block_size],
					                &val_buf[(indices[ind_row][ind_col] - 1) * block_size],
					                weights[ind_row][ind_col],
					                block_size);
				}
			}
		}

		TypedArray<double> out = factory.createArrayFromBuffer<double>(out_dims, std::move(out_buf));
		return out;
	}

public:
	/* External prototype:
		TypedArray<double> A = accumndarray(TypedArray<double> values,
											TypedArray<double> indices,
											TypedArray<double> weights);
		size(A) = v1 x ... x vk x max(indices)
		size(values) = v1 x ... x vk x h, where h >= max(indices)
		size(indices) = n x m, where n is the number of returned accumulants and m is the (max.) number of references to values
		size(weights) = n x m, as above (optional, interpreted as ones(size(indices)) if missing)

		min(indices) = 0, with 0 indicating accumulation of zero
		max(indices) = length(values)

		Example:
		Values = [[V1]
		          [V2]
		          [V3]
		          [V4]
		          [V5]
		          [V6]]
		Indices = [2 3;
		           1 0;
		           5 6;
		           3 2];
		Weights = [0.5 1;
		             1 0;
		             2 3;
		             0 1];
		->
		Output = [[0.5 V2 + V3]
		          [V1]
		          [2 V5 + 3 V6]
		          [0 V3 + 1 V2]]
	*/
	void operator()(ArgumentList outputs, ArgumentList inputs) {
		checkArgs(outputs, inputs);

		TypedArray<double> values = inputs[0];
		TypedArray<uint32_t> indices = doubleArrayToUints(inputs[1]);
		ArrayFactory factory;
		TypedArray<double> weights = factory.createArray<double>(indices.getDimensions());
		if (inputs.size() == 3) {
			weights = inputs[2];
		} else {
			std::fill(weights.begin(), weights.end(), 1.0);
		}

		TypedArray<double> sums = indexedAccumulate(values, indices, weights);
		outputs[0] = sums;
	}
};