%----------------------------------------------------------------- 
%
% Antti Hannukainen and Mika Juntunen 3.11.2005
%
% Vectorization by Antti Hannukainen 29.8.2007
%      Revision by Antti Hannukainen 27.10.2008
%      Expansion by Topi Kuutela 2020-2021
%
%-----------------------------------------------------------------
%
% Construct mesh - structure from a given triangulation. 
%
% The triangulation is expressed using matrices p and t. 
% 
% The p-matrix has dimension 2 x Np. Each column of p represents 
% coordinates of a single node of the mesh. For example p(:,i) 
% is coordinate of node i. Coordinates are assumed to be unique. 
% 
% The t-matrix has dimension 3xNt. Each column of t represents nodal 
% indeces of single triangle in the mesh. For example, t(:,i) is a vector [n1 n2 n3]' 
% such that p(:,n1), p(:,n2), and p(:,n3) are coordinates of the nodes of triangle i. 
%
% MESH STRUCTURE :
% 
%   mesh.p       = nodes in a 2xN-matrix
%
%   mesh.t       = triangles in a 3xN- matrix. 
% 
%   mesh.edges   = a matrix of all edges in the mesh. Each column is an edge 
%                  [n1 ; n2] where n1 and n2 are nodal indeces to p-matrix such that 
%                  n1 < n2; 
%
%   mesh.t2e     = a matrix connecting  triangle's and edges's. 
%                  Each column corresponds to single triangle in the mesh and 
%                  has triangle's edge indeces in the order n1->n2, n2->n3,
%                  n1->n3. 
%
%   mesh.e2t     = inverse of t2e. Boundary edges satisfy mesh.e2t(2,:) = 0.
%
%   mesh.bn      = list of all boundary nodes.
%   mesh.in      = list of all interor nodes.
%
%
% CALLING SYNTAX IS 
%
% function mesh = inittri(p,t)
%
%   p       = points matrix (described above)
%   t       = triangles matrix (described above)
%     
%   mesh    = trimesh structure corresponding to (p,t)
%


function mesh = inittri(p,t, boundary_indices)
    mesh.p  = p;
    mesh.t  = t;

    % Initalize size variables
    Nt = size(t,2);

    e = [1 2; 2 3; 1 3]';

    edges = [];

    for i=1:size(e,2)
      edges = [edges [ sort( [t(e(1,i),:); t(e(2,i),:)],1)]];
    end

    [mesh.edges,~,mesh.t2e] = sc_unique(edges);
    mesh.t2e = reshape(mesh.t2e,Nt,3)';

    % mesh.e2t
    e = [mesh.t2e(1,:) mesh.t2e(2,:)  mesh.t2e(3,:)];
    t = repmat([1:Nt],1,3);

    [ef,If] = unique(e, 'first');
    [el,Il] = unique(e, 'last');

    mesh.e2t(1,ef) = t(If);
    mesh.e2t(2,el) = t(Il);

    mesh.e2t(2, find((mesh.e2t(1,:)-mesh.e2t(2,:))==0))=0;

    bedgeids = find(mesh.e2t(2,:) == 0);
    mesh.be2t = mesh.e2t(1, bedgeids);
    mesh.bedges = mesh.edges(:, bedgeids);
    if nargin == 3
        mesh.bedgestream = boundary_stream(mesh, boundary_indices);
    else
        mesh.bedgestream = boundary_stream(mesh);
    end
    mesh.bedges = [mesh.bedgestream(:, 1), circshift(mesh.bedgestream(:, 1), [-1, 0])]';
    mesh.bn = unique([mesh.bedges']);
    mesh.in = setdiff(1:size(mesh.p,2),mesh.bn);

    [~, ~, ib1] = intersect(mesh.bn, mesh.t(1, :));
    [~, ~, ib2] = intersect(mesh.bn, mesh.t(2, :));
    [~, ~, ib3] = intersect(mesh.bn, mesh.t(3, :));
    mesh.bts = unique([ib1; ib2; ib3]);

    mesh.affine = affine_tri(mesh);
    mesh.Nt = size(mesh.t, 2);
    mesh.Np = size(mesh.p, 2);
    mesh.Ne = size(mesh.edges, 2);
    mesh.Nbe = size(mesh.bedges, 2);
    
    % Create p2t from mesh.t (= t2p). Columnwise store ts per p and then
    % squeeze extra zeros out.
    cols = []; rows = [];
    for tid = 1:mesh.Nt
        cols(end+1:end+3) = mesh.t(:, tid);
        rows(end+1:end+3) = tid;
    end
    p2t_sparse = sparse(rows, cols, rows, mesh.Nt, mesh.Np);
    [i, j, ~] = find(p2t_sparse);
    rows_per_col = full(sum(logical(p2t_sparse), 1));
    new_rows = [];
    for pid = 1:mesh.Np
        new_rows(end+1:end+rows_per_col(pid)) = 1:rows_per_col(pid);
    end
    length(new_rows)
    mesh.p2t = full(sparse(new_rows, j, i));
end

function b_stream = boundary_stream(mesh, boundary_indices)
    %       index         arc-length start      arc-length end
    % o b_stream(1, 1)    b_stream(1, 4) = 0.0
    % |
    % |
    % |
    % o b_stream(2, 1)    b_stream(2, 4),       b_stream(1, 5)
    % |
    % |
    % |
    % o b_stream(end, 1)  b_stream(end, 4),     b_stream(2, 5)
    % |
    % | 
    % |                    
    %                         <none>            b_stream(end, 5)
    % 
    %
    %
    % b_stream(:, 1) = mesh.p indices
    % b_stream(:, 2) = mesh.p total support length
    % b_stream(:, 3) = electrode index (for vertex, i.e. the corresponding edge shouldn't be included!)
    % b_stream(:, 4) = arc length parametrization edge start
    % b_stream(:, 5) = arc length parametrization edge end
    b_stream = zeros(size(mesh.bedges, 2), 2);
    bedge_pairs = sortrows([mesh.bedges'; fliplr(mesh.bedges')]);

    b_stream(1, 1) = bedge_pairs(1, 1);
    b_stream(2, 1) = bedge_pairs(1, 2);
    for step = 2:length(b_stream)-1
        current_node = b_stream(step, 1);
        next_candidates = bedge_pairs(bedge_pairs(:, 1) == current_node, 2);
        if (next_candidates(1) ~= b_stream(step-1))
            b_stream(step + 1, 1) = next_candidates(1);
        else
            b_stream(step + 1, 1) = next_candidates(2);
        end
    end

    if nargin == 2
        [~, I] = sort(b_stream(:, 1));
        b_stream(I, 3) = boundary_indices(:, 2);

        for rots = 0:size(b_stream, 1)
            temp = circshift(b_stream(:, 3), [rots, 0]);
            if temp(1) == 0 && temp(end) == 0
                b_stream = circshift(b_stream, [rots, 0]);
                break;
            end
            if rots == size(b_stream, 1)
                warning("Could not shift indexing so that no electrode is at the end or beginning.");
            end
        end
    end
    
    p1 = mesh.p(:, b_stream(:, 1));
    p2 = mesh.p(:, circshift(b_stream(:, 1), [-1, 0]));
    dist = sqrt(sum((p2 - p1).^2))';
    arcl = cumsum(dist);
    b_stream(:, 4) = [0; arcl(1:end-1)];
    b_stream(:, 5) = [arcl(1:end)];
    b_stream(:, 2) = dist;
    b_stream(:, 2) = b_stream(:, 2) + circshift(dist, [1, 0]);
    b_stream(:, 6) = circshift(b_stream(:, 1), [-1 0]);
end
